/*
modules.cc
Copyright 2007-2021 Matt Heinzen

This file is part of Color Machine.

Color Machine is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Color Machine is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Color Machine; if not, write to the Free Software Foundation, inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "modules.h"

#include "colormachinesdl.h"
#include "utils.h"

#include <math.h>


char const* getModuleName(int moduleIdx)
{
	switch (moduleIdx)
	{
	case MODULE_DIAMONDS: return "Diamonds";
	case MODULE_GRID: return "Grid";
	case MODULE_ORGANIC: return "Organic";
	case MODULE_SPIKES: return "Spikes";
	case MODULE_WORMS: return "Worms";
	case MODULE_PIPELINE: return "Pipeline";
	case MODULE_FRACTURED: return "Fractured";
	case MODULE_CHECKERBOARD: return "Checkerboard";
	case MODULE_MICROCHIP: return "Microchip";
	case MODULE_OPPOSITION: return "Opposition";
	case MODULE_INFINITY: return "Infinity";
	case MODULE_CONTOUR: return "Contour";
	case MODULE_VORTEX: return "Vortex";
	case MODULE_THORNS: return "Thorns";
	case MODULE_WAVES: return "Waves";
	case MODULE_NUCLEUS: return "Nucleus";
	case MODULE_JELLY: return "Jelly";
	case MODULE_RAYS: return "Rays";
	case MODULE_SPIRAL: return "Spiral";
	case MODULE_OSMOSIS: return "Osmosis";
	case MODULE_MULTIPOLARITY: return "Multipolarity";
	case MODULE_RIPPLES: return "Ripples";
	case MODULE_BEYOND: return "Beyond";
	case MODULE_CONTINUUM: return "Continuum";
	case MODULE_SPHERES: return "Spheres";
	case MODULE_GODS_ORB: return "God's Orb";
	default: return "Unknown";
	}
}

void drawModuleDiamonds()
{
	int nDiamond = randomRange(150) + 20;
	Uint8 colorIndex = randomColor();

	for (int i = 0; i < nDiamond; i++)
	{
		int size = randomInt(gSurfaceW * 0.03f, gSurfaceW * 0.15f);
		int x = randomRange(gSurfaceW);
		int y = randomRange(gSurfaceH);
		int colorDir = randomSign();

		if (randomBool())
		{
			for (int j = -size; j <= size; j++)
			{
				cycleColor(colorIndex, colorDir);
				drawVLine(x + j, y - abs(j) + size, y + abs(j) - size, colorIndex);
			}
		}
		else
		{
			for (int j = -size; j <= size; j++)
			{
				cycleColor(colorIndex, colorDir);
				drawHLine(x - abs(j) + size, x + abs(j) - size, y + j, colorIndex);
			}
		}
	}
}

void drawModuleGrid()
{
	Uint8 colorIndex = randomColor();
	int colorDir = randomSign();
	for (int i = 0; i < gSurfaceW; i += 2)
	{
		cycleColor(colorIndex, colorDir);
		drawVLine(i + 1, 0, gSurfaceH - 1, colorIndex);
	}

	colorIndex = randomColor();
	colorDir = randomSign();
	for (int i = 0; i < gSurfaceH; i += 2)
	{
		cycleColor(colorIndex, colorDir);
		drawHLine(0, gSurfaceW - 1, i + 1, colorIndex);
	}
}

void drawModuleOrganic()
{
	int cx = gSurfaceW / 2;
	int cy = gSurfaceH / 2;
	float radius = gSurfaceH * 0.25f;

	float range = gSurfaceW * 0.03f;

	int numNodes = randomInt(1000, 2500);

	Uint8 colorIndex = randomColor();
	int colorDir = randomSign();

	float oldx = randomFloat(cx - radius, cx + radius);
	float oldy = randomFloat(cy - radius, cy + radius);

	bool useMomentum = true; // much better than original, will probably make permanent
	if (useMomentum)
	{
		numNodes *= 4;
		range *= 0.06f;
	}
	float px = 0.0f;
	float py = 0.0f;

	for (int i = 0; i < numNodes; i++)
	{
		cycleColor(colorIndex, colorDir);

		float x, y;

		if (useMomentum)
		{
			x = oldx + randomFloat(-range, range) + px;
			y = oldy + randomFloat(-range, range) + py;

			px = (x - oldx) * 0.9f + (cx - x) * 0.0005f;
			py = (y - oldy) * 0.9f + (cy - y) * 0.0005f;
		}
		else
		{
			x = oldx + randomFloat(-range, range);
			y = oldy + randomFloat(-range, range);

			if ((cx - x) * (cx - x) + (cy - y) * (cy - y) > gSurfaceH * gSurfaceH / 4)
			{
				x -= (x - cx) / 4;
				y -= (y - cy) / 4;
			}
		}

		drawLine(oldx, oldy, x, y, colorIndex);
		drawLine(oldx - 1, oldy, x - 1, y, colorIndex);
		drawLine(oldx + 1, oldy, x + 1, y, colorIndex);
		drawLine(oldx, oldy - 1, x, y - 1, colorIndex);
		drawLine(oldx, oldy + 1, x, y + 1, colorIndex);

		oldx = x;
		oldy = y;
	}
}

void drawModuleSpikes()
{
	// TODO: 4 different directions?

	Shape shape = randomShape();
	bool round = randomBool();
	int length = gSurfaceW / 4;

	int dx = randomSign();
	int dy = randomSign();

	for (int i = 3; i > -2; i--)
	{
		for (int j = 2; j > -2; j--)
		{
			Uint8 colorIndex = randomColor();

			for (int k = 0; k < length; k++)
			{
				int x = i * length + k * 2;
				int y = j * length + k * 2;

				cycleColor(colorIndex);
				drawShape(shape, x, y, length - k, colorIndex);
			}
		}
	}
}

void drawModuleWorms()
{
	int numWorms = randomInt(gSurfaceW * 0.03f, gSurfaceW * 0.1f);
	int wormSize = randomInt(gSurfaceW * 0.02f, gSurfaceW * 0.03f);
	float persist = randomFloat(0.5f, 0.95f);

	for (int i = 0; i < numWorms; i++)
	{
		int l = randomInt(500, 1000);
		float x = randomRange(gSurfaceW);
		float y = randomRange(gSurfaceH);
		Uint8 colorIndex = randomColor();
		int colorDir = randomSign();

		float px = 0.0f;
		float py = 0.0f;

		for (int j = 0; j < l; j++)
		{
			cycleColor(colorIndex, colorDir);

			float dx = randomFloat(-1.0f, 1.0f) + px;
			float dy = randomFloat(-1.0f, 1.0f) + py;

			x += dx;
			y += dy;

			px = dx * persist;
			py = dy * persist;

			if (x < 0) px += 1.0f;
			if (y < 0) py += 1.0f;
			if (x > gSurfaceW - 10) px -= 1.0f;
			if (y > gSurfaceH - 10) py -= 1.0f;

			// make the worms start near full size, grow, and then taper towards the end drawn in front
			drawDisc(x, y, (int)(wormSize * sin(M_PI * (0.5 * j / l + 0.35f))), colorIndex);
		}
	}
}

void drawModulePipeline()
{
	Shape shape = randomShape();
	int numPipes = randomInt(8, 15);
	int pipeSize = randomInt(gSurfaceW * 0.015f, gSurfaceW * 0.030f);
	bool round = randomBool();

	for (int i = 0; i < numPipes; i++)
	{
		float x = randomRange(gSurfaceW);
		float y = randomRange(gSurfaceH);
		Uint8 colorIndex = randomColor();
		int colorDir = randomSign();

		while (x > -pipeSize && y > -pipeSize && x < gSurfaceW && y < gSurfaceH)
		{
			int l = randomInt(10, 40);
			float dx = randomFloat(-2, 2);
			float dy = randomFloat(-2, 2);

			for (int j = 0; j < l && x > -pipeSize && y > -pipeSize && x < gSurfaceW && y < gSurfaceH; j++)
			{
				x += dx;
				y += dy;

				cycleColor(colorIndex, colorDir);
				drawShape(shape, x, y, pipeSize, colorIndex);
			}
		}
	}
}

void drawModuleFractured()
{
	const int DIM = gSurfaceW / 6;
	const int STEP = gSurfaceW / 200;

	Shape shape = randomShape();
	int size = randomInt(gSurfaceW * 0.01f, gSurfaceW * 0.02f);
	int offsetRange = randomInt(gSurfaceW * 0.003f, gSurfaceW * 0.07f);

	Uint8 colorIndex = randomColor();
	int colorDir = randomSign();

	int flipI1, flipI2, flipJ1, flipJ2;
	do
	{
		flipI1 = randomBool() ? -1 : 1;
		flipI2 = randomBool() ? -1 : 1;
		flipJ1 = randomBool() ? -1 : 1;
		flipJ2 = randomBool() ? -1 : 1;
	} while (flipI1 * flipI2 * flipJ1 * flipJ2 == 1); // disallow combinations resulting in this bands

	for (int i = -DIM; i < DIM; i++)
	{
		cycleColor(colorIndex, colorDir);

		for (int j = -DIM; j < DIM; j++)
		{
			int dx = randomInt(-offsetRange, offsetRange);
			int dy = randomInt(-offsetRange, offsetRange);
			int x = flipI1 * i * STEP + flipJ1 * j * STEP + dx + gSurfaceW / 2;
			int y = flipI2 * i * STEP + flipJ2 * j * STEP + dy + gSurfaceH / 2;

			if (x > -size && y > -size && x < gSurfaceW && y < gSurfaceH)
				drawShape(shape, x, y, size / 2, colorIndex);
		}
	}
}

void drawModuleCheckerboard()
{
	int size = randomInt(gSurfaceW * 0.03f, gSurfaceW * 0.1f);

	int nStepX = gSurfaceW / size + 1;
	int nStepY = gSurfaceH / size + 1;
	int offX = (size - gSurfaceW % size) / 2;
	int offY = (size - gSurfaceH % size) / 2;

	for (int i = 0; i < nStepX; i++)
	{
		for (int j = 0; j < nStepY; j++)
		{
			Uint8 colorIndex = randomColor();
			int colorDir = randomSign();
			bool isVertical = (i + j) % 2;

			for (int k = 0; k < size; k++)
			{
				cycleColor(colorIndex, colorDir);

				if (isVertical) drawVLine(i * size + k - offX, j * size - offY, (j + 1) * size - 1 - offY, colorIndex);
				else drawHLine(i * size - offX, (i + 1) * size - 1 - offX, j * size + k - offY, colorIndex);
			}
		}
	}
}

void drawModuleMicrochip()
{
	int r = randomInt(3, 20);

	for (int y = 0; y < gSurfaceH; y++)
	{
		Uint8 colorIndex = randomColor();
		for (int x = 0; x < gSurfaceW; x++)
		{
			cycleColor(colorIndex, randomInt(-r, r));
			plotPixel(x, y, colorIndex);
		}
	}

	for (int x = 0; x < gSurfaceW; x += 2)
	{
		Uint8 colorIndex = randomColor();
		for (int y = 0; y < gSurfaceH; y++)
		{
			cycleColor(colorIndex, randomInt(-r, r));
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleOpposition()
{
	int x1, y1, x2, y2;
	double l;

	do
	{
		x1 = randomInt(gSurfaceW * 0.1f, gSurfaceW * 0.9f);
		y1 = randomInt(gSurfaceH * 0.1f, gSurfaceH * 0.9f);
		x2 = randomInt(gSurfaceW * 0.1f, gSurfaceW * 0.9f);
		y2 = randomInt(gSurfaceH * 0.1f, gSurfaceH * 0.9f);
		l = sqrt(abs(x1 - x2) * abs(x1 - x2) + abs(y1 - y2) * abs(y1 - y2));
	} while (l < gSurfaceH * 0.1f || l > gSurfaceH * 0.4f);

	for (int i = 0; i < gSurfaceW; i++)
	{
		for (int j = 0; j < gSurfaceH; j++)
		{
			int dx1 = abs(i - x1);
			int dy1 = abs(j - y1);
			int dx2 = abs(i - x2);
			int dy2 = abs(j - y2);
			double dist1 = sqrt(dx1 * dx1 + dy1 * dy1);
			double dist2 = sqrt(dx2 * dx2 + dy2 * dy2);
			Uint8 colorIndex = (Uint8)(128.0 + 127.0 * (dist1 - dist2) / l);
			plotPixel(i, j, colorIndex);
		}
	}
}

void drawModuleInfinity()
{
	int halfW = gSurfaceW / 2;
	int cx = gSurfaceW / 2;
	int cy = gSurfaceH / 2;

	float scale = randomFloat(1.0f, 10.0f) * gSurfaceW * 20.0f;

	int colorDir = randomSign();

	for (int i = halfW / 200 + 1; i < halfW * 1.4; i++)
	{
		float dist = scale / i;
		int diag = (int)(i * 0.707f);

		Uint8 colorIndex = int2Color(dist * colorDir);

		drawVLine(cx - i, cy - i, cy + i, colorIndex);
		drawVLine(cx + i, cy - i, cy + i, colorIndex);
		drawHLine(cx - i, cx + i, cy - i, colorIndex);
		drawHLine(cx - i, cx + i, cy + i, colorIndex);
		drawLine(cx - diag - diag, cy - diag + diag, cx - diag + diag, cy - diag - diag, colorIndex);
		drawLine(cx - diag - diag, cy + diag - diag, cx - diag + diag, cy + diag + diag, colorIndex);
		drawLine(cx + diag - diag, cy - diag - diag, cx + diag + diag, cy - diag + diag, colorIndex);
		drawLine(cx + diag - diag, cy + diag + diag, cx + diag + diag, cy + diag - diag, colorIndex);
	}
}

void drawModuleContour()
{
	const int MIN_POLES = 1;
	const int MAX_POLES = 8;

	int mode = randomRange(4);

	int numHigh = randomInt(MIN_POLES, MAX_POLES);
	int numLow = randomInt(MIN_POLES, MAX_POLES);

	if (mode == 0)
	{
		numHigh = MAX_POLES;
		numLow = MAX_POLES;
	}

	int xh[MAX_POLES], yh[MAX_POLES], xl[MAX_POLES], yl[MAX_POLES];
	for (int i = 0; i < MAX_POLES; i++)
	{
		xh[i] = randomRange(gSurfaceW);
		yh[i] = randomRange(gSurfaceH);
		xl[i] = randomRange(gSurfaceW);
		yl[i] = randomRange(gSurfaceH);
	}

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			// calculate pixel XY distances to high poles
			int dxh[MAX_POLES], dyh[MAX_POLES];
			for (int l = 0; l < numHigh; l++)
			{
				dxh[l] = abs(x - xh[l]);
				dyh[l] = abs(y - yh[l]);
			}

			// calculate pixel XY distances to low poles
			int dxl[MAX_POLES], dyl[MAX_POLES];
			for (int l = 0; l < numLow; l++)
			{
				dxl[l] = abs(x - xl[l]);
				dyl[l] = abs(y - yl[l]);
			}

			// calculate pixel linear distances to high and low poles
			double dhsq[MAX_POLES],  dlsq[MAX_POLES];
			for (int l = 0; l < numHigh; l++) dhsq[l] = dxh[l] * dxh[l] + dyh[l] * dyh[l];
			for (int l = 0; l < numLow; l++) dlsq[l] = dxl[l] * dxl[l] + dyl[l] * dyl[l];

			// calculate pixel distance to nearest high pole
			double dhNear = dhsq[0];
			for (int k = 1; k < numHigh; k++)
			{
				if (dhsq[k] < dhNear) dhNear = dhsq[k];
			}
			dhNear = sqrt(dhNear);

			// calculate pixel distance to nearest low pole
			double dlNear = dlsq[0];
			for (int k = 1; k < numLow; k++)
			{
				if (dlsq[k] < dlNear) dlNear = dlsq[k];
			}
			dlNear = sqrt(dlNear);

			if (dhNear < 1.0) dhNear = 1.0;
			if (dlNear < 1.0) dlNear = 1.0;

			// color ramp by relative distance to nearest high and low poles
			// TODO: base ramp multiples on gSurfaceW
			Uint8 colorIndex;
			switch (mode)
			{
			case 0: colorIndex = int2Color(int(128.0 + 0.001 * (dlNear * dlNear - dhNear * dhNear))); break;
			case 1: colorIndex = int2Color(int(128.0 + 1.3 * (dlNear - dhNear))); break;
			case 2: colorIndex = int2Color(int(128.0 + 13.0 * (sqrt(dlNear) - sqrt(dhNear)))); break;
			case 3: colorIndex = int2Color(int(128.0 + 100.0 * (sqrt(sqrt(dlNear)) - sqrt(sqrt(dhNear))))); break;
			}

			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleVortex()
{
	int cx = gSurfaceW / 2;
	int cy = gSurfaceH / 2;

	Shape shape = randomShape();
	double theta = 0.0;
	double thetaStep = randomSign() * 0.02;
	double colorstep = randomFloat(64.0f, 192.0f);

	float shapeMult = shape == SHAPE_DIAMOND ? 4.0f : 3.5f;

	for (int i = 0; ; i++)
	{
		theta += thetaStep;
		int r = i * i / 18432;
		int l = sqrt(r) * shapeMult;
		int x = cx + 1.33 * r * cos(theta);
		int y = cy - r * sin(theta);

		if (x > gSurfaceW + l && y > gSurfaceH + l) return;

		Uint8 colorIndex = int2Color(theta * colorstep);

		drawShape(shape, x, y, l, colorIndex);
	}
}

void drawModuleThorns()
{
	int pad = gSurfaceH / 20;
	int rangeX = gSurfaceW - pad * 2;
	int rangeY = gSurfaceH - pad * 2;
	int minThickness = gSurfaceW / 15;
	int rangeThickness = gSurfaceW / 10;

	int n = randomInt(15, 45);

	for (int i = 0; i < n; i++)
	{
		Uint8 colorIndex = randomColor();
		int x = randomInt(pad, pad + rangeX);
		int y = randomInt(pad, pad + rangeY);
		int l = randomInt(minThickness, minThickness + rangeThickness);

		// sides
		if (randomBool())
		{
			int x2 = randomBool() ? 0 : gSurfaceW - 1;
			int y2 = randomInt(pad, pad + rangeY);
			if (randomBool())
			{
				for (int j = y2; j < y2 + l; j++)
				{
					cycleColor(colorIndex);
					if (j < gSurfaceH) drawLine(x, y, x2, j, colorIndex);
				}
			}
			else
			{
				for (int j = y2; j > y2 - l; j--)
				{
					cycleColor(colorIndex);
					if (j >= 0) drawLine(x, y, x2, j, colorIndex);
				}
			}
		}
		// top/bottom
		else
		{
			int y2 = randomBool() ? 0 : gSurfaceH - 1;
			int x2 = randomInt(pad, pad + rangeX);
			if (randomBool())
			{
				for (int j = x2; j < x2 + l; j++)
				{
					cycleColor(colorIndex);
					if (j < gSurfaceW) drawLine(x, y, j, y2, colorIndex);
				}
			}
			else
			{
				for (int j = x2; j > x2 - l; j--)
				{
					cycleColor(colorIndex);
					if (j > -1) drawLine(x, y, j, y2, colorIndex);
				}
			}
		}
	}
}

void drawModuleWaves()
{
	double r = randomFloat(gSurfaceW * 0.005f, gSurfaceW * 0.025f);
	double f = randomBool() ? randomFloat(0.5f, 1.0f) : randomFloat(1.0f, 2.0f);

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			Uint8 colorIndex = int2Color(64.0 * (sin(x / r * f) + sin(y / r)));
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleNucleus()
{
	int cx = randomInt(gSurfaceW * 0.2f, gSurfaceW * 0.8f);
	int cy = randomInt(gSurfaceH * 0.2f, gSurfaceH * 0.8f);
	double r = randomFloat(5.0f, 25.0f);
	double f = randomBool() ? randomFloat(0.4f, 1.0f) : randomFloat(1.0f, 2.5f);
	int n = randomRange(5) + randomRange(3) + randomRange(2) + 2;

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			double a = hypot(abs(x - cx), abs(y - cy));

			Uint8 colorIndex = int2Color(64.0 * sin(x / r * f) + 64.0 * sin(y / r) + a * n);
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleJelly()
{
	double multY = 1.0 / randomFloat(5.0f, 25.0f);
	double multX = 1.0 / randomFloat(5.0f, 25.0f);
	
	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			Uint8 colorIndex = int2Color(32.0 * (
				sin(x * multX) + sin(x * multX * 1.5) + sin(y * multY) + sin(y * multY * 1.5)));
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleRays()
{
	float cx = randomFloat(gSurfaceW * 0.25f, gSurfaceW * 0.75f);
	float cy = randomFloat(gSurfaceH * 0.25f, gSurfaceH * 0.75f);
	double period = M_PI * 2.0 / randomInt(3, 11);

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			double theta = atan2(x - cx, y - cy);
			Uint8 colorIndex = int2Color(theta / period * 255.0);
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleSpiral()
{
	int cx = randomRange(gSurfaceW / 2) + gSurfaceW / 4;
	int cy = randomRange(gSurfaceH / 2) + gSurfaceH / 4;

	double flip = randomSign();
	double f = M_PI * 2.0 / (double)(randomRange(5) + 2);
	double n = (randomRange(20) + 1) / 5.0;

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			int dx = x - cx;
			int dy = cy - y;
			double a = hypot(dx, dy);
			double theta = flip * atan2((double)dy, (double)dx);

			Uint8 colorIndex = int2Color((int)(theta * 255 / f + a * n));
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleOsmosis()
{
	// ORIGINAL USING INTEGER MATH
	/**
	int colorIndex;
	int r, x, x2, y, y2, i, j, d, l;

	r = randomRange(8000 / gSurfaceW) + 2000 / gSurfaceW;
	x = randomRange(gSurfaceW / 2) + gSurfaceW / 4;
	x2 = randomRange(gSurfaceW / 2) + gSurfaceW / 4;
	y = randomRange(gSurfaceH / 2) + gSurfaceH / 4;
	y2 = randomRange(gSurfaceH / 2) + gSurfaceH / 4;

	for (i = 0; i < gSurfaceW; i++)
	{
		for (j = 0; j < gSurfaceH; j++)
		{
			d = (int)(hypot(i - x, j - y));
			l = (int)(hypot(i - x2, j - y2));
			colorIndex = int2Color(d * l * r);

			plotPixel(i, j, colorIndex);
		}
	}
	/**/

	// EXPERIMENTAL CONVERSION TO FP MATH
	// gives very different results, similar to ???
	// NOTE: d and l being ints before multiply give characteristic effect
	// TODO: try to get like old behavior but smooth like new behavior
	/**/
	double r = 0.1 * (randomRange(20) + 1);
	int x1 = randomRange(gSurfaceW / 2) + gSurfaceW / 4;
	int x2 = randomRange(gSurfaceW / 2) + gSurfaceW / 4;
	int y1 = randomRange(gSurfaceH / 2) + gSurfaceH / 4;
	int y2 = randomRange(gSurfaceH / 2) + gSurfaceH / 4;

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			double a1 = hypot(x - x1, y - y1);
			double a2 = hypot(x - x2, y - y2);
			
			Uint8 colorIndex;

			colorIndex = int2Color((int)a1 * (int)a2 * r); // old behavior, cooler but heavy aliasing, a1 and a2 being ints before mult give characteristic effect
			//colorIndex = int2Color(round(a1) * round(a2) * r); // same as original, including aliasing
			//colorIndex = int2Color(a1 * a2 * r); // new behavior, very smooth, just not as cool
			//colorIndex = int2Color((int)(a1 * 10.0) * (int)(a2 * 10.0) * r / 100.0); // worst of both worlds, less cool effect but with aliasing
			//colorIndex = int2Color((int)(a1 * 0.1) * (int)(a2 * 0.1) * r * 100.0); // disco ball aliasing, but maybe keep as a mode

			// TODO: maybe try super sampling the integer method 2x2 for each final pixel? not working well
			if (x > gSurfaceW / 2)
			{
				/*a1 = 0.25 * hypot(x - 0.25 - x1, y - 0.25 - y1)
					+ 0.25 * hypot(x + 0.25 - x1, y - 0.25 - y1)
					+ 0.25 * hypot(x - 0.25 - x1, y + 0.25 - y1)
					+ 0.25 * hypot(x + 0.25 - x1, y + 0.25 - y1);
				a2 = 0.25 * hypot(x - 0.25 - x2, y - 0.25 - y2)
					+ 0.25 * hypot(x + 0.25 - x2, y - 0.25 - y2)
					+ 0.25 * hypot(x - 0.25 - x2, y + 0.25 - y2)
					+ 0.25 * hypot(x + 0.25 - x2, y + 0.25 - y2);
				colorIndex = int2Color(a1 * a2 * r);*/
				
				/*a1 = hypot(x - 0.25 - x1, y - 0.25 - y1);
				a2 = hypot(x - 0.25 - x2, y - 0.25 - y2);
				int c = (int)((int)a1 * (int)a2 * r);
				a1 = hypot(x + 0.25 - x1, y - 0.25 - y1);
				a2 = hypot(x + 0.25 - x2, y - 0.25 - y2);
				c += (int)((int)a1 * (int)a2 * r);
				a1 = hypot(x - 0.25 - x1, y + 0.25 - y1);
				a2 = hypot(x - 0.25 - x2, y + 0.25 - y2);
				c += (int)((int)a1 * (int)a2 * r);
				a1 = hypot(x + 0.25 - x1, y + 0.25 - y1);
				a2 = hypot(x + 0.25 - x2, y + 0.25 - y2);
				c += (int)((int)a1 * (int)a2 * r);
				colorIndex = int2Color(c / 4);*/
			}
			

			plotPixel(x, y, colorIndex);
		}
	}
	/**/
}

void drawModuleMultipolarity()
{
	int useCosine = randomRange(3); // favor using cosine

	int x1 = randomRange(gSurfaceW);
	int y1 = randomRange(gSurfaceH);
	int x2 = randomRange(gSurfaceW);
	int y2 = randomRange(gSurfaceH);
	int x3 = randomRange(gSurfaceW);
	int y3 = randomRange(gSurfaceH);
	double r = randomFloat(gSurfaceW * 0.03f, gSurfaceW * 0.2f);

	double cx, cy;
	if (randomBool())
	{
		cx = (x1 + x2 + x3) / 3.0f;
		cy = (y1 + y2 + y3) / 3.0f;
	}
	else
	{
		cx = randomRange(gSurfaceW);
		cy = randomRange(gSurfaceH);
	}

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			double a1 = hypot(x - x1, y - y1);
			double a2 = hypot(x - x2, y - y2);
			double a3 = hypot(x - x3, y - y3);
			double asum = a1 + a2 + a3;
			double sinHypotR = sin(hypot(x - cx, y - cy) / r);

			Uint8 colorIndex = useCosine == 0
				? colorIndex = int2Color(asum * sinHypotR)
				: colorIndex = int2Color(asum * cos(sinHypotR));
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleRipples()
{
	int x1 = randomRange(gSurfaceW);
	int y1 = randomRange(gSurfaceH);
	int x2 = randomRange(gSurfaceW);
	int y2 = randomRange(gSurfaceH);
	float r = randomFloat(0.75f, 0.75f + gSurfaceW * 0.02f);

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			int dx1 = x - x1;
			int dx2 = x - x2;
			int dy1 = y - y1;
			int dy2 = y - y2;
			int sumDistSqr = dx1 * dx1 + dy1 * dy1 + dx2 * dx2 + dy2 * dy2;
			float f = sumDistSqr / r;

			Uint8 colorIndex = int2Color(f);
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleBeyond()
{
	int x1 = randomRange(gSurfaceW);
	int y1 = randomRange(gSurfaceH);
	int x2 = randomRange(gSurfaceW);
	int y2 = randomRange(gSurfaceH);
	float r = randomFloat(0.8f, 0.8f + gSurfaceW * 0.015f);

	for (int x = 0; x < gSurfaceW; x++)
	{
		for(int y = 0; y < gSurfaceH; y++)
		{
			int dx1 = x - x1;
			int dx2 = x - x2;
			int dy1 = y - y1;
			int dy2 = y - y2;
			int distSqrX = dx1 * dx1 + dx2 * dx2;
			int distSqrY = dy1 * dy1 + dy2 * dy2;
			float fx = distSqrX / r;
			float fy = distSqrY / r;

			Uint8 colorIndex = int2Color(fx - fy);
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleContinuum()
{
	float x1 = randomFloat(-0.5f * gSurfaceW, 1.5f * gSurfaceW);
	float y1 = randomFloat(-0.5f * gSurfaceH, 1.5f * gSurfaceH);
	float x2 = randomFloat(0.0f, gSurfaceW);
	float y2 = randomFloat(0.0f, gSurfaceH);
	float r = randomFloat(0.25f * gSurfaceW, 8.0f * gSurfaceW);

	for (int x = 0; x < gSurfaceW; x++)
	{
		for (int y = 0; y < gSurfaceH; y++)
		{
			float dx1 = x - x1;
			float dx2 = x - x2;
			float dy1 = y - y1;
			float dy2 = y - y2;
			float dxy1 = x - y1;
			float dxy2 = x - y1;
			float distCubeX = dx1 * dx1 * dx1 + dx2 * dx2 * dx2;
			float distCubeYYXY = dy1 * dy1 * dxy1 + dy2 * dy2 * dxy2; // this started out as copy-paste typos, but is much better this way!
			float fx = distCubeX / r;
			float fyyxy = distCubeYYXY / r;

			Uint8 colorIndex = int2Color(fx + fyyxy);
			plotPixel(x, y, colorIndex);
		}
	}
}

void drawModuleSpheres()
{
	int n = randomInt(10, 25);
	for (int k = 0; k < n; k++)
	{
		int cMin = randomRange(255);
		int cRange = randomSign() * (randomRange(512) + 128);
		int cMax = cMin + cRange;
		int cAvg = (cMax + cMin) / 2;

		int x = randomRange(gSurfaceW);
		int y = randomRange(gSurfaceH);
		int r = randomRange(gSurfaceW / 6) + gSurfaceW / 50;

		double theta = randomRange(360) * M_PI / 180.0;
		double focaldist = randomRange(r) * 0.8;
		int x2 = (int)(x + focaldist * cos(theta));
		int y2 = (int)(y + focaldist * sin(theta));

		for (int i = x - r; i <= x + r; i++)
		{
			for (int j = y - r; j <= y + r; j++)
			{
				double d = hypot(x - i, y - j);

				if (d <= r)
				{
					double edgedist = 1.75 * r - d;
					double focaldist = hypot(x2 - i, y2 - j);

					Uint8 colorIndex = int2Color(cMin + cRange * (focaldist - edgedist) / (focaldist + edgedist));
					plotPixel(i, j, colorIndex);
				}
			}
		}
	}
}

void drawModuleGodsOrb()
{
	int mode = randomRange(4) + 1;
	/**mode = 1;/**/

	int cx = gSurfaceW / 2;
	int cy = gSurfaceH / 2;
	double r = gSurfaceH * 0.425f;
	double off = gSurfaceW / 20;
	double off2 = gSurfaceW / 10;

	// display the background
	for (double x2 = -gSurfaceW / 2; x2 < gSurfaceW / 2; x2 += 3)
	{
		for (double y2 = -gSurfaceH / 2; y2 < gSurfaceH / 2; y2 += 3)
		{
			int x = (int)x2 + cx;
			int y = (int)y2 + cy;

			Uint8 colorIndex;
			switch (mode)
			{
			case 1:
				colorIndex = int2Color(y % 2 == 0 ? x : y);
				break;
			case 2:
				colorIndex = int2Color((sin(x2 / 8.0) + sin(y2 / 8.0)) * 64.0);
				break;
			case 3:
				if (x % 2 == 1)
				{
					double xlo = x2 - off2;
					double ylo = y2 - off2;
					double yhi = y2 + off2;
					if (y % 2 == 1) colorIndex = int2Color(3.0 * sqrt(xlo * xlo + ylo * ylo));
					else colorIndex = int2Color(3.0 * sqrt(xlo * xlo + yhi * yhi));
				}
				else
				{
					double xhi = x2 + off2;
					double ylo = y2 - off2;
					double yhi = y2 + off2;
					if (y % 2 == 1) colorIndex = int2Color(3.0 * sqrt(xhi * xhi + ylo * ylo));
					else colorIndex = int2Color(3.0 * sqrt(xhi * xhi + yhi * yhi));
				}
				break;
			case 4:
				colorIndex = int2Color(atan(y2 / x2) * 255.0 / M_PI_2 + sqrt(x2 * x2 + y2 * y2) * 6.0);
				break;
			}

			plotPixel(x, y, colorIndex);
		}
	}
	
	const double epsilon = 1.01;
	double sinHalfPiEpsilon = sin(M_PI_2 * epsilon);
	double cosHalfPiEpsilon = cos(M_PI_2 * epsilon);

	// display the sphere
	for (int i = (int)(cx - r); i < cx + r; i++)
	{
		for (int j = (int)(cy - r); j < cy + r; j++)
		{
			double focaldist = hypot((cx - i), (cy - j));
			if (focaldist <= r)
			{
				double dx = i - cx;
				double dy = j - cy;
				
				double theta = atan2(dy, dx);
				double edgedist = r - focaldist;

				// coordinates based on integration of
				// d = 25 / sin(M_PI_2 * (r * 1.01 - float(focaldist)) / r);
				double rFocalRatio = (r * epsilon - focaldist) / r;
				double rFocalRatioHalfPi = M_PI_2 * rFocalRatio;
				double d = -30.0 * log(fabs(2500.0 / sin(rFocalRatioHalfPi) - 2500.0 * cos(rFocalRatioHalfPi) / sin(rFocalRatioHalfPi)));
				d += 30.0 * log(fabs(2500.0 / sinHalfPiEpsilon - 2500.0 * cosHalfPiEpsilon / sinHalfPiEpsilon));

				// calculate texturemap coordinates using d
				double x2 = d * cos(theta);
				double y2 = d * sin(theta);

				Uint8 colorIndex;
				switch (mode)
				{
				case 1:
					colorIndex = int2Color((int)(x2 * 4.0) % 2 == 0 ? x2 * 4 : y2 * 4);
					break;
				case 2:
					colorIndex = int2Color((sin(x2 / 8.0) + sin(y2 / 8.0)) * 64.0);
					break;
				case 3:
					if (i % 2)
					{
						double xlo = x2 - off;
						double ylo = y2 - off;
						double yhi = y2 + off;
						if (j % 2) colorIndex = int2Color(7.0 * sqrt(xlo * xlo + ylo * ylo));
						else colorIndex = int2Color(7.0 * sqrt(xlo * xlo + yhi * yhi));
					}
					else
					{
						double xhi = x2 + off;
						double ylo = y2 - off;
						double yhi = y2 + off;
						if (j % 2) colorIndex = int2Color(7.0 * sqrt(xhi * xhi + ylo * ylo));
						else colorIndex = int2Color(7.0 * sqrt(xhi * xhi + yhi * yhi));
					}
					break;
				case 4:
					colorIndex = int2Color(atan(y2 / x2) * 255.0 / M_PI_2 + sqrt(x2 * x2 + y2 * y2) * 6.0);
					break;
				}
				
				plotPixel(i, j, colorIndex);
			}
		}
	}
}

void drawModule(int moduleIdx)
{
	switch (moduleIdx)
	{
	case MODULE_DIAMONDS: drawModuleDiamonds(); break;
	case MODULE_GRID: drawModuleGrid(); break;
	case MODULE_ORGANIC: drawModuleOrganic(); break;
	case MODULE_SPIKES: drawModuleSpikes(); break;
	case MODULE_WORMS: drawModuleWorms(); break;
	case MODULE_PIPELINE: drawModulePipeline(); break;
	case MODULE_FRACTURED: drawModuleFractured(); break;
	case MODULE_CHECKERBOARD: drawModuleCheckerboard(); break;
	case MODULE_MICROCHIP: drawModuleMicrochip(); break;
	case MODULE_OPPOSITION: drawModuleOpposition(); break;
	case MODULE_INFINITY: drawModuleInfinity(); break;
	case MODULE_CONTOUR: drawModuleContour(); break;
	case MODULE_VORTEX: drawModuleVortex(); break;
	case MODULE_THORNS: drawModuleThorns(); break;
	case MODULE_WAVES: drawModuleWaves(); break;
	case MODULE_NUCLEUS: drawModuleNucleus(); break;
	case MODULE_JELLY: drawModuleJelly(); break;
	case MODULE_RAYS: drawModuleRays(); break;
	case MODULE_SPIRAL: drawModuleSpiral(); break;
	case MODULE_OSMOSIS: drawModuleOsmosis(); break;
	case MODULE_MULTIPOLARITY: drawModuleMultipolarity(); break;
	case MODULE_RIPPLES: drawModuleRipples(); break;
	case MODULE_BEYOND: drawModuleBeyond(); break;
	case MODULE_CONTINUUM: drawModuleContinuum(); break;
	case MODULE_SPHERES: drawModuleSpheres(); break;
	case MODULE_GODS_ORB: drawModuleGodsOrb(); break;
	//case 27: drawModuleOrion(); break;
	//case 28: drawModule(); break;
	//case 29: drawModule(); break;
	default: drawModuleGodsOrb(); break;
	}
}

