/*
palettes.cc
Copyright 2007-2021 Matt Heinzen

This file is part of Color Machine.

Color Machine is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

Color Machine is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Color Machine; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "palettes.h"

#include "utils.h"


char const* getPaletteName(int paletteIdx)
{
	switch (paletteIdx)
	{
	case PALETTE_RED_PULSE: return "Red Pulse";
	case PALETTE_GREEN_PULSE: return "Green Pulse";
	case PALETTE_BLUE_PULSE: return "Blue Pulse";
	case PALETTE_YELLOW_PULSE: return "Yellow Pulse";
	case PALETTE_CYAN_PULSE: return "Cyan Pulse";
	case PALETTE_MAGENTA_PULSE: return "Magenta Pulse";
	case PALETTE_RAINBOW: return "Rainbow";
	case PALETTE_BROWNIAN: return "Brownian";
	default: return "Unknown";
	}
}

void setPaletteColor(SDL_Color *palette, int idx, float red, float green, float blue)
{
	// set the color for the specified palette index

	palette[idx].r = (Uint8)(red * 255.0f);
	palette[idx].g = (Uint8)(green * 255.0f);
	palette[idx].b = (Uint8)(blue * 255.0f);
	palette[idx].a = 255;
}

void initPaletteBlack(SDL_Color *palette)
{
	for (int i = 0; i < 256; i++) setPaletteColor(palette, i, 0.0f, 0.0f, 0.0f);
}


void copyPalette(SDL_Color *src, SDL_Color *dest)
{
	memcpy(dest, src, 256 * sizeof(SDL_Color));
}

void rotatePalette(SDL_Color *palette, bool const forward)
{
	// rotate the palette by one position in the specified direction

	SDL_Color endColor;

	if (forward)
	{
		endColor = palette[255];
		for (int i = 255; i > 1; i--) palette[i] = palette[i - 1];
		palette[1] = endColor;
	}
	else
	{
		endColor = palette[1];
		for (int i = 1; i < 255; i++) palette[i] = palette[i + 1];
		palette[255] = endColor;
	}
}

void rotateRed(SDL_Color *palette, bool const forward)
{
	// rotate the palette's red components by one position in the specified direction

	SDL_Color endColor;

	if (forward)
	{
		endColor.r = palette[255].r;
		for (int i = 255; i > 1; i--) palette[i].r = palette[i - 1].r;
		palette[1].r = endColor.r;
	}
	else
	{
		endColor.r = palette[1].r;
		for (int i = 1; i < 255; i++) palette[i].r = palette[i + 1].r;
		palette[255].r = endColor.r;
	}
}

void rotateGreen(SDL_Color *palette, bool const forward)
{
	// rotate the palette's green components by one position in the specified direction

	SDL_Color endColor;

	if (forward)
	{
		endColor.g = palette[255].g;
		for (int i = 255; i > 1; i--) palette[i].g = palette[i - 1].g;
		palette[1].g = endColor.g;
	}
	else
	{
		endColor.g = palette[1].g;
		for (int i = 1; i < 255; i++) palette[i].g = palette[i + 1].g;
		palette[255].g = endColor.g;
	}
}

void rotateBlue(SDL_Color *palette, bool const forward)
{
	// rotate the palette's blue components by one position in the specified direction

	SDL_Color endColor;

	if (forward)
	{
		endColor.b = palette[255].b;
		for (int i = 255; i > 1; i--) palette[i].b = palette[i - 1].b;
		palette[1].b = endColor.b;
	}
	else
	{
		endColor.b = palette[1].b;
		for (int i = 1; i < 255; i++) palette[i].b = palette[i + 1].b;
		palette[255].b = endColor.b;
	}
}

void fadePalette_x256(SDL_Color *palette, int const numSteps)
{
	for (int i = 0; i < 256; i++)
	{
		if (palette[i].r > numSteps) palette[i].r -= numSteps;
		else palette[i].r = 0;
		if (palette[i].g > numSteps) palette[i].g -= numSteps;
		else palette[i].g = 0;
		if (palette[i].b > numSteps) palette[i].b -= numSteps;
		else palette[i].b = 0;
	}
}

void calcPulseLow(bool linear, int idx, float &primary, float &secondary)
{
	if (linear)
	{
		primary = idx < 64 ? idx / 64.0f : 1.0f;
		secondary = idx < 64 ? 0.0f : (idx - 64) / 64.0f;
	}
	else
	{
		/**
		// smooth but no pure hue
		primary = sin(idx * M_PI / 256.0);
		secondary = idx < 64 ? 0.0f : sin((idx - 64) * M_PI / 128.0);
		/**/
		// smooth and pure hue, too soft?
		primary = idx < 64 ? sin(idx * M_PI / 128.0) : 1.0f;
		secondary = idx < 64 ? 0.0f : sin((idx - 64) * M_PI / 128.0);
		/**/
	}
}

void initPaletteRedPulse(SDL_Color *palette, bool linear)
{
	// map palette values for "red Pulse" color set

	setPaletteColor(palette, 0, 0.0f, 0.0f, 0.0f);
	for (int i = 0; i < 128; i++)
	{
		float r, gb;
		calcPulseLow(linear, i, r, gb);
		setPaletteColor(palette, i + 1, r, gb, gb);
		setPaletteColor(palette, 255 - i, r, gb, gb);
	}
}

void initPaletteGreenPulse(SDL_Color *palette, bool linear)
{
	// map palette values for "green Pulse" color set

	setPaletteColor(palette, 0, 0.0f, 0.0f, 0.0f);
	for (int i = 0; i < 128; i++)
	{
		float g, rb;
		calcPulseLow(linear, i, g, rb);
		setPaletteColor(palette, i + 1, rb, g, rb);
		setPaletteColor(palette, 255 - i, rb, g, rb);
	}
}

void initPaletteBluePulse(SDL_Color *palette, bool linear)
{
	// map palette values for "blue Pulse" color set

	setPaletteColor(palette, 0, 0.0f, 0.0f, 0.0f);
	for (int i = 0; i < 128; i++)
	{
		float b, rg;
		calcPulseLow(linear, i, b, rg);
		setPaletteColor(palette, i + 1, rg, rg, b);
		setPaletteColor(palette, 255 - i, rg, rg, b);
	}
}

void initPaletteYellowPulse(SDL_Color *palette, bool linear)
{
	// map palette values for "Yellow Pulse" color set

	setPaletteColor(palette, 0, 0.0f, 0.0f, 0.0f);
	for (int i = 0; i < 128; i++)
	{
		float rg, b;
		calcPulseLow(linear, i, rg, b);
		setPaletteColor(palette, i + 1, rg, rg, b);
		setPaletteColor(palette, 255 - i, rg, rg, b);
	}
}

void initPaletteCyanPulse(SDL_Color *palette, bool linear)
{
	// map palette values for "Cyan Pulse" color set

	setPaletteColor(palette, 0, 0.0f, 0.0f, 0.0f);
	for (int i = 0; i < 128; i++)
	{
		float gb, r;
		calcPulseLow(linear, i, gb, r);
		setPaletteColor(palette, i + 1, r, gb, gb);
		setPaletteColor(palette, 255 - i, r, gb, gb);
	}
}

void initPaletteMagentaPulse(SDL_Color *palette, bool linear)
{
	// map palette values for "Magenta Pulse" color set

	setPaletteColor(palette, 0, 0.0f, 0.0f, 0.0f);
	for (int i = 0; i < 128; i++)
	{
		float rb, g;
		calcPulseLow(linear, i, rb, g);
		setPaletteColor(palette, i + 1, rb, g, rb);
		setPaletteColor(palette, 255 - i, rb, g, rb);
	}
}

void initPaletteRainbow(SDL_Color *palette, bool linear)
{
	// map palette values for "Rainbow" color set

	setPaletteColor(palette, 0, 0.0f, 0.0f, 0.0f);

	if (linear)
	{
		for (int i = 1; i <= 42; i++) setPaletteColor(palette, i, 1.0f, i / 42.0f, 0.0f);				// red-yellow
		for (int i = 43; i <= 85; i++) setPaletteColor(palette, i, (85 - i) / 43.0f, 1.0f, 0.0f);		// yellow-green
		for (int i = 86; i <= 128; i++) setPaletteColor(palette, i, 0.0f, 1.0f, (i - 85) / 43.0f);		// green-cyan
		for (int i = 129; i <= 171; i++) setPaletteColor(palette, i, 0.0f, (171 - i) / 43.0f, 1.0f);	// cyan-blue
		for (int i = 172; i <= 213; i++) setPaletteColor(palette, i, (i - 171) / 42.0f, 0.0f, 1.0f);	// blue-magenta
		for (int i = 214; i <= 255; i++) setPaletteColor(palette, i, 1.0f, 0.0f, (255 - i) / 42.0f);	// magenta-red
	}
	else
	{
		for (int i = 1; i <= 42; i++) setPaletteColor(palette, i, 1.0f, sin(i * M_PI_2 / 42.0f), 0.0f);				// red-yellow
		for (int i = 43; i <= 85; i++) setPaletteColor(palette, i, sin((85 - i) * M_PI_2 / 43.0f), 1.0f, 0.0f);		// yellow-green
		for (int i = 86; i <= 128; i++) setPaletteColor(palette, i, 0.0f, 1.0f, sin((i - 85) * M_PI_2 / 43.0f));	// green-cyan
		for (int i = 129; i <= 171; i++) setPaletteColor(palette, i, 0.0f, sin((171 - i) * M_PI_2 / 43.0f), 1.0f);	// cyan-blue
		for (int i = 172; i <= 213; i++) setPaletteColor(palette, i, sin((i - 171) * M_PI_2 / 42.0f), 0.0f, 1.0f); 	// blue-magenta
		for (int i = 214; i <= 255; i++) setPaletteColor(palette, i, 1.0f, 0.0f, sin((255 - i) * M_PI_2 / 42.0f)); 	// magenta-red
	}
}

// NOTE: large range to update ratio makes smoother palettes
// best values: 0.05 & 0.025 (smooth), 0.05 & 0.005 (sedate), 0.07 & 0.007 (moderate),
// 0.1 & 0.005 (moderately wild), 0.05 & 0.01 (wild), 0.1 & 0.01 (very wild)
#define COLOR_DELTA_RANGE 0.05
#define COLOR_DELTA_UPDATE 0.0025

void initPaletteBrownian(SDL_Color *palette)
{
	// map palette values for "Brownian" color set

	double r = randomRange(1000) / 1000.0 * 1.0;
	double g = randomRange(1000) / 1000.0 * 1.0;
	double b = randomRange(1000) / 1000.0 * 1.0;
	double dr = randomRange(100) / 100.0 * COLOR_DELTA_RANGE - COLOR_DELTA_RANGE / 2.0;
	double dg = randomRange(100) / 100.0 * COLOR_DELTA_RANGE - COLOR_DELTA_RANGE / 2.0;
	double db = randomRange(100) / 100.0 * COLOR_DELTA_RANGE - COLOR_DELTA_RANGE / 2.0;

	for (int i = 0; i < 128; i++) {
		setPaletteColor(palette, i, r, g, b);
		setPaletteColor(palette, 255 - i, r, g, b);

		dr += randomRange(100) / 100.0 * COLOR_DELTA_UPDATE - COLOR_DELTA_UPDATE / 2.0;
		dg += randomRange(100) / 100.0 * COLOR_DELTA_UPDATE - COLOR_DELTA_UPDATE / 2.0;
		db += randomRange(100) / 100.0 * COLOR_DELTA_UPDATE - COLOR_DELTA_UPDATE / 2.0;

		if (dr < -COLOR_DELTA_RANGE / 2.0) dr += COLOR_DELTA_UPDATE / 2.0;
		if (dr > COLOR_DELTA_RANGE / 2.0) dr -= COLOR_DELTA_UPDATE / 2.0;
		if (dg < -COLOR_DELTA_RANGE / 2.0) dg += COLOR_DELTA_UPDATE / 2.0;
		if (dg > COLOR_DELTA_RANGE / 2.0) dg -= COLOR_DELTA_UPDATE / 2.0;
		if (db < -COLOR_DELTA_RANGE / 2.0) db += COLOR_DELTA_UPDATE / 2.0;
		if (db > COLOR_DELTA_RANGE / 2.0) db -= COLOR_DELTA_UPDATE / 2.0;

		r += dr;
		g += dg;
		b += db;

		if (r < 0.0)
		{
			r = 0.0;
			dr = -dr;
		}
		else if (r > 1.0)
		{
			r = 1.0;
			dr = -dr;
		}

		if (g < 0.0)
		{
			g = 0.0;
			dg = -dg;
		}
		else if (g > 1.0)
		{
			g = 1.0;
			dg = -dg;
		}

		if (b < 0.0)
		{
			b = 0.0;
			db = -db;
		}
		else if (b > 1.0)
		{
			b = 1.0;
			db = -db;
		}
	}

	setPaletteColor(palette, 0, 0.0f, 0.0f, 0.0f);
}

void initPalette(SDL_Color *palette, int paletteIdx, bool linearPalette)
{
	switch (paletteIdx)
	{
	case PALETTE_RED_PULSE: initPaletteRedPulse(palette, linearPalette); break;
	case PALETTE_GREEN_PULSE: initPaletteGreenPulse(palette, linearPalette); break;
	case PALETTE_BLUE_PULSE: initPaletteBluePulse(palette, linearPalette); break;
	case PALETTE_YELLOW_PULSE: initPaletteYellowPulse(palette, linearPalette); break;
	case PALETTE_CYAN_PULSE: initPaletteCyanPulse(palette, linearPalette); break;
	case PALETTE_MAGENTA_PULSE: initPaletteMagentaPulse(palette, linearPalette); break;
	case PALETTE_RAINBOW: initPaletteRainbow(palette, linearPalette); break;
	case PALETTE_BROWNIAN: initPaletteBrownian(palette); break;
	default: initPaletteRainbow(palette, linearPalette); break;
	}
}

